<?php
/**
 * @package     CSVI
 * @subpackage  File
 *
 * @author      RolandD Cyber Produksi <contact@rolandd.com>
 * @copyright   Copyright (C) 2006 - [year] RolandD Cyber Produksi. All rights reserved.
 * @license     GNU/GPLv3 http://www.gnu.org/licenses/gpl-3.0.html
 * @link        https://rolandd.com
 */

use Joomla\CMS\Language\Text;

defined('_JEXEC') or die;

/**
 * XML file importer.
 *
 * @package     CSVI
 * @subpackage  File
 * @since       6.0
 */
class CsviHelperFileImportXml extends CsviHelperFile
{
	/**
	 * Contains the data that is read from file
	 *
	 * @var    XMLReader
	 * @since  3.0
	 */
	protected $data;

	/**
	 * Contains the record start to read from the XML
	 *
	 * @var    string
	 * @since  3.0
	 */
	private $recordContainer = '';

	/**
	 * Contains the record name to read from the XML
	 *
	 * @var    string
	 * @since  3.0
	 */
	private $recordName = '';

	/**
	 * Contains the record name to read from the XML prior to 8.0.0
	 *
	 * @var    string
	 * @since  8.0.0
	 */
	private $oldRecordName = '';

	/**
	 * Internal line pointer
	 *
	 * @var    int
	 * @since  3.0
	 */
	public $linepointer = 0;

	/**
	 * Holds the CSVI fields helper
	 *
	 * @var    CsviHelperImportFields
	 * @since  6.0
	 */
	protected $fields;

	/**
	 * Node tree
	 *
	 * @var    array
	 * @since  7.20.0
	 */
	private $nodeTree = [];

	/**
	 * Data tree
	 *
	 * @var    array
	 * @since  7.20.0
	 */
	private $dataTree = [];

	/**
	 * Set to count the lines in an XML file.
	 *
	 * @var    bool
	 *
	 * @since   7.20.0
	 */
	private $countOnly = false;

	/**
	 * Track the parent levels that have been read
	 *
	 * @var    array
	 * @since  7.21.0
	 */
	private $parents = [];

	/**
	 * A list of nodes from the container node
	 *
	 * A container node is the node containing all the subrecords
	 *
	 * @var    array
	 * @since  7.21.0
	 */
	private $containerNodes = [];

	/**
	 * A list of data entries that correspond to the keys in the container node
	 *
	 * @var    array
	 * @since  7.21.0
	 */
	private $containerData = [];

	/**
	 * Keep track of the current read nodes
	 *
	 * @var    array
	 * @since  7.21.0
	 */
	private $recordNodes = [];

	/**
	 * The data corresponding to the open conunter
	 *
	 * @var    array
	 * @since  7.21.0
	 */
	private $recordData = [];

	/**
	 * Tracks the number of rows that have been processed
	 *
	 * @var    integer
	 * @since  7.21.0
	 */
	private $lineCounter = 0;

	/**
	 * Hold the list of containers
	 *
	 * @var    array
	 * @since  7.21.0
	 */
	private $containers = [];

	/**
	 * Hold the list of records
	 *
	 * @var    array
	 * @since  7.21.0
	 */
	private $records = [];

	/**
	 * Set if we started a new container
	 *
	 * @var    boolean
	 * @since  7.21.0
	 */
	private $containerOpen = false;

	/**
	 * Keep track of the active record being requested to be read
	 *
	 * @var    integer
	 * @since  7.21.0
	 */
	private $activeRecord = -1;

	/**
	 * Stores the relation between a container and a record. This way we know which container belongs to which record.
	 *
	 * @var    array
	 * @since  7.21.0
	 */
	private $containerRecordMap = [];

	/**
	 * Count the node we are currently processing
	 *
	 * @var    integer
	 * @since  7.21.0
	 */
	private $nodeCounter = 0;

	/**
	 * Set if we changed the path set in the template to a nested path
	 *
	 * @var    boolean
	 * @since  7.21.0
	 */
	private $convertedPath = false;

	/**
	 * Tracks the number of rows that have been processed
	 *
	 * @var    integer
	 * @since  8.1.0
	 */
	private $recordCounter = 0;

	/**
	 * Load the column headers from a file.
	 *
	 * @return   mixed    array when column headers are found | false if column headers cannot be read.
	 *
	 * @throws  Exception
	 * @since   3.0
	 */
	public function loadColumnHeaders()
	{
		$columnheaders = [];
		$continue      = true;
		$line          = 0;

		// Make sure the file is loaded
		$this->openFile();

		// Start reading the XML file
		while ($this->data->read())
		{
			// Only read the chosen records
			if ($this->data->nodeType == XMLREADER::ELEMENT
				&& $this->data->name == $this->recordName
				&& $continue
			)
			{
				// Start reading the record
				while ($this->data->read() && $continue)
				{
					switch ($this->data->nodeType)
					{
						case (XMLREADER::ELEMENT):
							// Check if it has attributes
							if ($this->data->hasAttributes)
							{
								$parent[] = $this->data->name;

								// Get the attributes
								while ($this->data->moveToNextAttribute())
								{
									// The attribute name
									if (empty($parent))
									{
										$field_name = $this->data->name;
									}
									else
									{
										$field_name = implode('/', $parent) . '/' . $this->data->name;
									}

									$columnheaders[] = $field_name;
								}
							}
							elseif (!$this->data->isEmptyElement)
							{
								$parent[] = $this->data->name;
							}
							break;
						case (XMLREADER::END_ELEMENT):
							$line++;
							array_pop($parent);

							if ($this->data->name == $this->recordName)
							{
								$continue = false;
							}
							break;
						case XMLReader::TEXT:
						case XMLReader::CDATA:
							// The field name
							if (empty($parent))
							{
								$field_name = $this->data->name;
							}
							else
							{
								$field_name = implode('/', $parent);
							}

							$columnheaders[] = $field_name;
							break;
					}
				}
			}
			elseif (!$continue)
			{
				break;
			}
		}

		// Reset the internal pointer
		$this->rewind();

		return $columnheaders;
	}

	/**
	 * Get the file position.
	 *
	 * @return  int  Current position in the file.
	 *
	 * @since   3.0
	 */
	public function getFilePos()
	{
		return $this->linepointer;
	}

	/**
	 * Set the file position
	 *
	 * To be able to set the file position correctly, the XML reader needs to be at the start of the file.
	 *
	 * @param   int  $pos  The position to move to.
	 *
	 * @return  int  Current position in the file.
	 *
	 * @since   3.0
	 */
	public function setFilePos($pos)
	{
		// Close the XML reader
		$this->closeFile(false);

		// Open a new XML reader
		$this->processFile();

		return $this->skipXmlRecords($pos);
	}

	/**
	 * Close the file.
	 *
	 * @param   bool  $removeFolder  Set if the folder should be removed.
	 *
	 * @return  void.
	 *
	 * @since   3.0
	 */
	public function closeFile($removeFolder = true)
	{
		if ($this->data->close())
		{
			$this->fp = false;
		}

		$this->closed = true;

		parent::closeFile($removeFolder);
	}

	/**
	 * Read the next line but determine which type of line to read
	 *
	 * @return  boolean True if data is found | False otherwise
	 *
	 * @since   7.20.0
	 */
	public function readNextLine(): bool
	{
		$this->activeRecord++;

		if (!strstr($this->recordName, '/'))
		{
			$this->findRootPath();
			$this->rewind();
		}

		return $this->readLine();
	}

	/**
	 * Find the root path of a given record name.
	 *
	 * @return  void
	 *
	 * @since   7.21.0
	 */
	private function findRootPath(): void
	{
		$path = [];

		while ($this->data->read())
		{
			if ($this->data->nodeType === XMLReader::ELEMENT && $this->data->name !== $this->recordName)
			{
				$path[] = $this->data->name;
			}

			if ($this->data->nodeType === XMLReader::END_ELEMENT)
			{
				$keys = array_keys($path, $this->data->name);
				$key  = end($keys);
				$path = array_slice($path, 0, $key);
			}

			if ($this->data->name === $this->recordName)
			{
				$path[] = $this->data->name;
				$this->template->set('xml_record_container', implode('/', $path));
				$this->template->set('xml_record_name', implode('/', $path));
				$this->convertedPath = true;
				$this->oldRecordName = $this->recordName;
				break;
			}
		}
	}

	/**
	 * Read an XML file line by line.
	 *
	 * @return  boolean True if data is found | False otherwise
	 *
	 * @since   7.21.0
	 */
	private function readLine(): bool
	{
		if (!isset($this->records[$this->activeRecord]))
		{
			$this->lineCounter = 0;

			while ($read = $this->data->read())
			{
				$parentKeys = array_keys($this->parents);
				$depth      = end($parentKeys);

				switch ($this->data->nodeType)
				{
					case XMLReader::ELEMENT:
						if ($depth === $this->data->depth)
						{
							$this->closeElement();
						}

						$this->openElement();
						break;
					case XMLReader::END_ELEMENT:
						$this->closeElement();
						break;
					case XMLReader::TEXT:
					case XMLReader::CDATA:
						$this->dataElement();
						break;
				}

				if ($this->lineCounter === 1 && !$this->countOnly)
				{
					break;
				}
			}

			if (!$read && !isset($this->records[$this->activeRecord]))
			{
				return false;
			}
		}

		$this->fillTree();
		$this->prepareFields();

		return $this->dataTree !== [];
	}

	/**
	 * Handle an XMLReader::ELEMENT.
	 *
	 * @return  void
	 *
	 * @since   7.21.0
	 */
	private function openElement(): void
	{
		$this->parents[] = $this->data->name;
		$this->nodeCounter++;

		if ($this->isContainer())
		{
			$this->containerOpen = true;
		}

		if ($this->data->hasAttributes)
		{
			$this->processAttributes();

			if ($this->isRecordNode() && !$this->isRecord())
			{
				$this->recordData[$this->nodeCounter] = '';
			}
			elseif ($this->isContainerNode() && !$this->isContainer())
			{
				$this->containerData[$this->nodeCounter] = '';
			}
		}

		if ($this->data->isEmptyElement)
		{
			$this->emptyElement();
		}
	}

	/**
	 * Handle an XMLReader::END_ELEMENT.
	 *
	 * @return  void
	 *
	 * @since   7.21.0
	 */
	private function closeElement(): void
	{
		$recordDepth = substr_count($this->recordContainer, '/');

		if (!isset($this->recordNodes[$this->nodeCounter]) && $this->isRecordNode() && !$this->isRecord())
		{
			$this->recordNodes[$this->nodeCounter] = $this->getCurrentPath();
		}
		elseif (!isset($this->containerNodes[$this->nodeCounter]) && !$this->isRecordNode() && $this->isContainerNode() && !$this->isContainer())
		{
			$this->containerNodes[$this->nodeCounter] = $this->getCurrentPath();
		}

		if ($this->isRecordNode() && !isset($this->recordData[$this->nodeCounter]))
		{
			$this->recordData[$this->nodeCounter] = '';
		}

		if ($this->isContainerNode() && !isset($this->containerData[$this->nodeCounter]))
		{
			$this->containerData[$this->nodeCounter] = '';
		}

		if ($this->data->depth >= $recordDepth)
		{
			if ($this->isRecord())
			{
				$this->fillRecords();
				$this->linepointer++;

				if ($this->isContainer())
				{
					$this->lineCounter++;
				}
			}
			elseif ($this->isContainer() && $this->containerOpen)
			{
				if (!isset($this->records[$this->linepointer]))
				{
					$this->fillRecords();
					$this->linepointer++;
				}

				$this->fillContainers();
				$this->containerOpen = false;
				$this->lineCounter++;
			}
		}

		$keys          = array_keys($this->parents, $this->data->name);
		$key           = end($keys);
		$this->parents = array_slice($this->parents, 0, $key);
	}

	/**
	 * Handle an XMLReader::TEXT and XMLReader::CDATA.
	 *
	 * @return  void
	 *
	 * @since   7.21.0
	 */
	private function emptyElement(): void
	{
		$recordDepth = count(explode('/', $this->recordContainer));

		if ($this->data->depth >= $recordDepth)
		{
			if ($this->isRecordNode())
			{
				$this->recordNodes[$this->nodeCounter] = $this->getCurrentPath();
				$this->recordData[$this->nodeCounter]  = '';
			}
			elseif ($this->isContainerNode())
			{
				$this->containerNodes[$this->nodeCounter] = $this->getCurrentPath();
				$this->containerData[$this->nodeCounter]  = '';
			}
		}

		$this->closeElement();
	}

	/**
	 * Handle an XMLReader::TEXT and XMLReader::CDATA.
	 *
	 * @return  void
	 *
	 * @since   7.21.0
	 */
	private function dataElement(): void
	{
		$recordDepth = count(explode('/', $this->recordContainer));

		if ($this->data->depth >= $recordDepth)
		{
			if ($this->isRecordNode())
			{
				if (!isset($this->recordData[$this->nodeCounter]))
				{
					$this->recordData[$this->nodeCounter] = '';
				}

				$this->recordNodes[$this->nodeCounter] = $this->getCurrentPath();
				$this->recordData[$this->nodeCounter]  .= $this->data->value;
			}
			elseif ($this->isContainerNode())
			{
				if (!isset($this->containerData[$this->nodeCounter]))
				{
					$this->containerData[$this->nodeCounter] = '';
				}

				$this->containerNodes[$this->nodeCounter] = $this->getCurrentPath();
				$this->containerData[$this->nodeCounter]  .= $this->data->value;
			}
		}
	}

	/**
	 * Get the current path in the XML tree where we are.
	 *
	 * @return  string  The path / delimited.
	 *
	 * @since   7.21.0
	 */
	private function getCurrentPath(): string
	{
		return implode('/', $this->parents);
	}

	/**
	 * Check if we are in a container node.
	 *
	 * @return  boolean  True if in a container node | False otherwise.
	 *
	 * @since   7.21.0
	 */
	private function isContainerNode(): bool
	{
		return !($this->recordContainer === '') && strstr($this->getCurrentPath(), $this->recordContainer) !== false;
	}

	/**
	 * Check if we are in a data node.
	 *
	 * @return  boolean  True if in a data node | False otherwise.
	 *
	 * @since   7.21.0
	 */
	private function isRecordNode(): bool
	{
		return strstr($this->getCurrentPath(), $this->recordName) !== false;
	}

	/**
	 * Check if the current path matches the container path.
	 *
	 * @return  boolean  True if in a container node | False otherwise.
	 *
	 * @since   7.21.0
	 */
	private function isContainer(): bool
	{
		return $this->getCurrentPath() === $this->recordContainer;
	}

	/**
	 * Check if the current path matches the record path.
	 *
	 * @return  boolean  True if in a container node | False otherwise.
	 *
	 * @since   7.21.0
	 */
	private function isRecord(): bool
	{
		return $this->getCurrentPath() === $this->recordName;
	}

	/**
	 * Fill the records array with the current read record
	 *
	 * @return  void
	 *
	 * @since   7.21.0
	 */
	private function fillRecords(): void
	{
		$record = [
			'nodes' => $this->recordNodes,
			'data'  => $this->recordData,
		];

		if ($this->countOnly && count($this->recordData) > 0)
		{
			$this->recordCounter++;
		}

		$this->records[$this->linepointer]            = $record;
		$this->recordNodes                            = [];
		$this->recordData                             = [];
		$this->containerRecordMap[$this->linepointer] = $this->activeRecord;
	}

	/**
	 * Fill the containers array with the current read container
	 *
	 * @return  void
	 *
	 * @since   7.21.0
	 */
	private function fillContainers(): void
	{
		$container                             = [
			'nodes' => $this->containerNodes,
			'data'  => $this->containerData,
		];
		$this->containers[$this->activeRecord] = $container;
		$this->containerNodes                  = [];
		$this->containerData                   = [];
	}

	/**
	 * Fill the tree with the nodes and data from both the container and records
	 *
	 * @return  void
	 *
	 * @since   7.21.0
	 */
	private function fillTree()
	{
		$this->nodeTree = $this->mergeTree(
			(isset($this->containerRecordMap[$this->activeRecord])
				? $this->containers[$this->containerRecordMap[$this->activeRecord]]['nodes'] ?? [] : []),
			($this->records[$this->activeRecord]['nodes'] ?? [])
		);
		$this->dataTree = $this->mergeTree(
			(isset($this->containerRecordMap[$this->activeRecord])
				? $this->containers[$this->containerRecordMap[$this->activeRecord]]['data'] ?? [] : []),
			($this->records[$this->activeRecord]['data'] ?? [])
		);
		unset($this->records[$this->activeRecord]);
	}

	/**
	 * Merge trees together to create a complete record.
	 *
	 * @param   array  $old  The old tree nodes
	 * @param   array  $new  The new tree nodes to add
	 *
	 * @return  array  The merged trees.
	 *
	 * @since   8.0.0
	 */
	private function mergeTree(array $old, array $new): array
	{
		foreach ($old as $key => $value)
		{
			$old[$key] = $new[$key] ?? $value;
		}

		foreach ($new as $key => $value)
		{
			if (!isset($old[$key]))
			{
				$old[$key] = $value;
			}
		}

		ksort($old);

		return $old;
	}

	/**
	 * Process the attributes of a node.
	 *
	 * @return  void
	 *
	 * @since   7.21.0
	 */
	private function processAttributes(): void
	{
		$nodeField = '';
		$dataField = '';

		if ($this->isRecordNode())
		{
			$nodeField = 'recordNodes';
			$dataField = 'recordData';
		}
		elseif ($this->isContainerNode())
		{
			$nodeField = 'containerNodes';
			$dataField = 'containerData';
		}

		if (!$nodeField || !$dataField)
		{
			return;
		}

		while ($this->data->moveToNextAttribute())
		{
			$this->{$nodeField}[$this->nodeCounter] = $this->getCurrentPath() . '/' . $this->data->name;;
			$this->{$dataField}[$this->nodeCounter] = $this->data->value;

			$this->nodeCounter++;
		}
	}

	/**
	 * Prepare the fields with import data.
	 *
	 * @return  void
	 *
	 * @since   7.21.0
	 */
	private function prepareFields()
	{
		if ($this->fields === null)
		{
			return;
		}

        $this->fields->reset();

		$existingFields = [];
		$fieldCount     = 0;

		foreach ($this->nodeTree as $key => $fieldName)
		{
			if (is_array($this->dataTree[$key]))
			{
				foreach ($this->dataTree[$key] as $counter => $value)
				{
					if ($this->convertedPath)
					{
						$fieldName = $this->getOldPath($fieldName);
					}

					$this->fields->set($fieldName, $value, $counter);
				}
			}
			else
			{
				$existingFields[] = $fieldName;

				if (in_array($fieldName, $existingFields))
				{
					$fieldTotalCount = array_count_values($existingFields);
					$fieldCount      = $fieldTotalCount[$fieldName];
				}

				if ($this->convertedPath)
				{
					$fieldName = $this->getOldPath($fieldName);
				}

				$this->fields->set($fieldName, $this->dataTree[$key], $fieldCount);
			}
		}
	}

	/**
	 * Get the name of an XML node pre 8.0.0.
	 *
	 * @return  string  The field name.
	 *
	 * @since   8.0.0
	 */
	private function getOldPath(string $fieldName): string
	{
		$names = explode('/', $fieldName);

		foreach ($names as $name)
		{
			array_shift($names);

			if ($name === $this->oldRecordName)
			{
				break;
			}
		}

		return implode('/', $names);
	}

	/**
	 * Open the file to read. We use XMLReader to support streaming to lower memory usage
	 *
	 * @return  bool  Returns true.
	 *
	 * @throws  Exception
	 * @since   3.0
	 */
	public function openFile()
	{
		if (!$this->fp)
		{
			$this->data = new XMLReader;
			$this->fp   = $this->data->open($this->filename);

			if ($this->fp === false)
			{
				$this->log->addStats('incorrect', JText::_('COM_CSVI_ERROR_XML_READING_FILE'));
				$app = JFactory::getApplication();
				$app->enqueueMessage(JText::_('COM_CSVI_ERROR_XML_READING_FILE'), 'error');

				return false;
			}
		}

		return true;
	}

	/**
	 * Process the file to import.
	 *
	 * @return  bool  Always returns true.
	 *
	 * @since   3.0
	 */
	public function processFile()
	{
		if ($this->openFile())
		{
			$this->recordContainer = $this->template->get('xml_record_container', '');
			$this->recordName      = $this->template->get('xml_record_name', '');
			$this->parents         = [];

			return true;
		}

		return false;
	}

	/**
	 * Sets the file pointer back to beginning.
	 *
	 * @return  void.
	 *
	 * @since   3.0
	 */
	public function rewind()
	{
		$this->linepointer = 0;

		// Close the file, so we can start reading from the beginning
		if ($this->data->close())
		{
			$this->fp = false;
		}

		$this->processFile();
	}

	/**
	 * Advances the file pointer 1 forward.
	 *
	 * Read one line and discard it
	 *
	 * @param   bool  $preview  True if called from preview.
	 *
	 * @return  void.
	 *
	 * @since   3.0
	 */
	public function next($preview = false)
	{
		if (!$preview)
		{
			$this->readNextLine();
		}
	}

	/**
	 * Skips through the XML file until the the required number 'record' nodes has been read
	 * Assume the file pointer is at the start of file.
	 *
	 * @param   int  $pos  The number of records to skip.
	 *
	 * @return  bool  True if records are skipped | false if records are not skipped.
	 *
	 * @since   3.0
	 */
	private function skipXmlRecords($pos)
	{
		$this->log->add('Forwarding to position: ' . $pos, false);

		// Check whether the pointer needs to be moved
		if ($pos <= 0)
		{
			return true;
		}

		while ($this->readNextLine())
		{
			if ($this->linepointer === (int) $pos)
			{
				return true;
			}
		}

		// Hit EOF before skipping the required number of records
		return false;
	}

	/**
	 * Returns the number of lines.
	 *
	 * @return  int  Returns the number of lines in the file.
	 *
	 * @since   6.0
	 */
	public function linecount()
	{
		$count = 0;

		if ($this->fp)
		{
			$currentPosition     = $this->getFilePos();
			$currentActiveRecord = $this->activeRecord = 0;
			$currentRecords      = $this->records = [];
			$this->recordCounter = 0;

			$this->rewind();

			$this->countOnly = true;

			$this->readNextLine();

			$count              = $this->recordCounter;
			$this->linepointer  = 0;
			$this->activeRecord = $currentActiveRecord;
			$this->records      = $currentRecords;
			$this->setFilePos($currentPosition);
		}

		return $count;
	}

	/**
	 * Return the node tree.
	 *
	 * @return  array  The list of nodes.
	 *
	 * @since   7.20.0
	 */
	public function getNodeTree(): array
	{
		return $this->nodeTree;
	}

	/**
	 * Return the data tree.
	 *
	 * @return  array  The list of values.
	 *
	 * @since   7.20.0
	 */
	public function getDataTree(): array
	{
		return $this->dataTree;
	}

	/**
	 * Return the details from XML file.
	 *
	 * @return  array  The details from XML file.
	 *
	 * @since   8.0.0
	 */
	public function getFileDetails()
	{
		return [
			Text::_('COM_CSVI_JFORM_XML_RECORD_CONTAINER_LABEL') => $this->recordContainer,
			Text::_('COM_CSVI_JFORM_XML_RECORD_NAME_LABEL')      => $this->oldRecordName !== '' ? $this->oldRecordName
				: $this->recordName,
		];
	}
}
